/* $Id: readcfg.c,v 1.2 1996/12/14 07:09:20 mff Exp mff $ */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>

#include "readcfg.h"
#include "tools.h"
#include "ftnaddr.h"

const char wsep[] = " \t\n\r\f";
typedef void (*meth) (char *s);

static char *path;
static int line;

int Loglevel = 4;
unsigned short iport = DEF_PORT;
unsigned short oport = DEF_PORT;
int call_delay = 60;
int rescan_delay = 60;
int nettimeout = DEF_TIMEOUT;
int oblksize = DEF_BLKSIZE;
int max_servers = 100;
int max_clients = 100;
char sysname[MAXSYSTEMNAME + 1] = "";
char sysop[MAXSYSOPNAME + 1] = "";
char location[MAXLOCATIONNAME + 1] = "";
char nodeinfo[MAXNODEINFO + 1] = "";
char inbound[MAXPATHLEN + 1] = ".";
char inbound_nonsecure[MAXPATHLEN + 1] = "";
FILE *Logfile = 0;

int nAddr = 0;
FTN_ADDR *pAddr = 0;

FTN_DOMAIN *pDomains = 0;

int nNod = 0;
FTN_NODE *pNod = 0;

void akas ()
{
  int i;
  char buf[80];
  FTN_DOMAIN *curr_domain;

  printf ("addr:");
  for (i = 0; i < nAddr; ++i)
  {
    ftnaddress_to_str (buf, pAddr + i);
    printf (" %s", buf);
  }
  printf ("\n");
  for (curr_domain = pDomains; curr_domain; curr_domain = curr_domain->next)
  {
    if (curr_domain->alias4 == 0)
      printf ("`%s`, `%s`, `%s`.\n",
	      curr_domain->name,
	      curr_domain->path,
	      curr_domain->dir);
    else
      printf ("`%s` alias for `%s`.\n",
	      curr_domain->name,
	      curr_domain->alias4->name);
  }
  printf ("\n");
  for (i = 0; i < nNod; ++i)
  {
    char b[FTN_ADDR_SZ + 1];

    ftnaddress_to_str (b, &pNod[i].fa);
    printf ("%s %s %s %s\n", b, pNod[i].host, pNod[i].pwd, pNod[i].filebox);
  }
  printf ("\n");
}

static void LogFile (char *s)
{
  char *w;

  if ((w = getword (s, 1)) != 0 && getword (s, 2) == 0)
  {
    if ((Logfile = fopen (w, "a")) == 0)
      Log (0, "%s: cannot open: %s", w, strerror (errno));
    else
      fputc ('\n', Logfile);
    free (w);
  }
  else
    Log (0, "%s: %i: the syntax is incorrect", path, line);
}

static void Inbound (char *s)
{
  char *w;

  if ((w = getword (s, 1)) != 0 && getword (s, 2) == 0)
  {
    strnzcpy (inbound, w, sizeof (inbound));
    free (w);

    for (w = inbound + strlen (inbound) - 1;
	 (*w == '/' || *w == '\\') && w >= inbound; --w)
      *w = 0;
  }
  else
    Log (0, "%s: %i: the syntax is incorrect", path, line);
}

static void InboundNonsecure (char *s)
{
  char *w;

  if ((w = getword (s, 1)) != 0 && getword (s, 2) == 0)
  {
    strnzcpy (inbound_nonsecure, w, sizeof (inbound_nonsecure));
    free (w);

    for (w = inbound_nonsecure + strlen (inbound_nonsecure) - 1;
	 (*w == '/' || *w == '\\') && w >= inbound_nonsecure; --w)
      *w = 0;
  }
  else
    Log (0, "%s: %i: the syntax is incorrect", path, line);
}

static void Sysname (char *s)
{
  char *w;

  if ((w = getword (s, 1)) != 0 && getword (s, 2) == 0)
  {
    strnzcpy (sysname, w, sizeof (sysname));
    free (w);
  }
  else
    Log (0, "%s: %i: the syntax is incorrect", path, line);
}

static void Sysop (char *s)
{
  char *w;

  if ((w = getword (s, 1)) != 0 && getword (s, 2) == 0)
  {
    strnzcpy (sysop, w, sizeof (sysop));
    free (w);
  }
  else
    Log (0, "%s: %i: the syntax is incorrectn", path, line);
}

static void Location (char *s)
{
  char *w;

  if ((w = getword (s, 1)) != 0 && getword (s, 2) == 0)
  {
    strnzcpy (location, w, sizeof (location));
    free (w);
  }
  else
    Log (0, "%s: %i: the syntax is incorrect", path, line);
}

static void Nodeinfo (char *s)
{
  char *w;

  if ((w = getword (s, 1)) != 0 && getword (s, 2) == 0)
  {
    strnzcpy (nodeinfo, w, sizeof (nodeinfo));
    free (w);
  }
  else
    Log (0, "%s: %i: the syntax is incorrect", path, line);
}

static void LogLevel (char *s)
{
  if ((Loglevel = atoi (s)) < 0)
    Log (0, "%s: %i: %i: incorrect Loglevel", path, line, Loglevel);
}

static void Iport (char *s)
{
  if ((iport = atoi (s)) <= 0)
    Log (0, "%s: %i: %i: incorrect port", path, line, iport);
}

static void Oport (char *s)
{
  if ((oport = atoi (s)) <= 0)
    Log (0, "%s: %i: %i: incorrect port", path, line, oport);
}

static void MaxServers (char *s)
{
  if ((max_servers = atoi (s)) <= 0)
    Log (0, "%s: %i: %i: incorrect value", path, line, max_servers);
}

static void MaxClients (char *s)
{
  if ((max_clients = atoi (s)) <= 0)
    Log (0, "%s: %i: %i: incorrect value", path, line, max_clients);
}

static void OBlkSize (char *s)
{
  if ((oblksize = atoi (s)) < MIN_BLKSIZE || oblksize > MAX_BLKSIZE)
    Log (0, "%s: %i: %i: incorrect block size", path, line, oblksize);
}

static void RescanDelay (char *s)
{
  if ((rescan_delay = atoi (s)) <= 0)
    Log (0, "%s: %i: %i: incorrect rescan-delay", path, line, rescan_delay);
}

static void CallDelay (char *s)
{
  if ((call_delay = atoi (s)) <= 0)
    Log (0, "%s: %i: %i: incorrect call-delay", path, line, call_delay);
}

static void Timeout (char *s)
{
  if ((nettimeout = atoi (s)) <= 0)
    Log (0, "%s: %i: %i: invalid timeout", path, line, nettimeout);
}

static void Domain (char *s)
{
/*  int i; */
  char *w1 = getword (s, 1);
  char *w2 = getword (s, 2);
  char *w3 = getword (s, 3);
  FTN_DOMAIN *new_domain;

  if (!w1 || !w2 || !w3)
    Log (0, "%s: %i: domain: not enough args", path, line);

  if (get_domain_info (w1) == 0)
  {
    new_domain = malloc (sizeof (FTN_DOMAIN));
    assert (new_domain != 0);

    strnzcpy (new_domain->name, w1, sizeof (new_domain->name));

    if (!stricmp (w2, "alias-for"))
    {
      FTN_DOMAIN *tmp_domain;

      if ((tmp_domain = get_domain_info (w3)) == 0)
	Log (0, "%s: %i: %s: undefined domain", path, line, w3);
      else
	new_domain->alias4 = tmp_domain;
      free (w2);
    }
    else
    {
      char *s;
      int z;

      if ((z = atoi (w3)) <= 0)
	Log (0, "%s: %i: invalid zone", path, line);

      new_domain->z = malloc (sizeof (int) * 2);

      assert (new_domain->z != 0);
      new_domain->z[0] = z;
      new_domain->z[1] = 0;
      new_domain->alias4 = 0;

      for (s = w2 + strlen (w2) - 1; (*s == '/' || *s == '\\') && s >= w2; --s)
	*s = 0;
      if ((s = max (strrchr (w2, '\\'), strrchr (w2, '/'))) == 0)
      {
	new_domain->dir = w2;
	new_domain->path = strdup (".");
      }
      else
      {
	new_domain->dir = strdup (s + 1);
	for (; *s == '/' || *s == '\\'; --s)
	  *s = 0;
	new_domain->path = w2;
      }
      if (strchr (new_domain->dir, '.'))
	Log (0, "%s: %i: there should be no extension for "
	     "the base outbound name", path, line);
    }
    new_domain->next = pDomains;
    pDomains = new_domain;
  }
  else
  {
    Log (0, "%s: %i: %s: duplicate domain", path, line, w1);
  }
  free (w1);
  free (w3);
}

static void Address (char *s)
{
  int i;
  char *w;

  for (i = 1; (w = getword (s, i)) != 0; ++i)
  {
    pAddr = realloc (pAddr, sizeof (FTN_ADDR) * (nAddr + 1));
    assert (pAddr != 0);
    if (!parse_ftnaddress (w, pAddr + nAddr))
      Log (0, "%s: %i: %s: the address cannot be parsed", path, line, w);
    if (!is4D (pAddr + nAddr))
      Log (0, "%s: %i: %s: must be at least a 4D address", path, line, w);
    if (!pAddr[nAddr].domain[0])
    {
      if (!pDomains)
	Log (0, "%s: %i: at least one domain must be defined first", path, line);
      strcpy (pAddr[nAddr].domain, get_def_domain ()->name);
    }
    ++nAddr;
    free (w);
  }
}

static void Node (char *s)
{
  char *w[4];
  int i;

  for (i = 0; i < 4; ++i)
    w[i] = getword (s, i + 1);

  if (getword (s, 5) || !w[0] || !w[1] || !w[2])
    Log (0, "%s: %i: the syntax is incorrect", path, line);

  pNod = realloc (pNod, sizeof (FTN_NODE) * (nNod + 1));
  assert (pNod != 0);

  if (!parse_ftnaddress (w[0], &pNod[nNod].fa))
    Log (0, "%s: %i: %s: the address cannot be parsed", path, line, w[0]);
  else
    exp_ftnaddress (&pNod[nNod].fa);
  if (get_node_info (&pNod[nNod].fa))
    Log (0, "%s: %i: %s: duplicate node entry", path, line, w[0]);

  strnzcpy (pNod[nNod].host, w[1], sizeof (pNod[nNod].host));
  strnzcpy (pNod[nNod].pwd, w[2], sizeof (pNod[nNod].pwd));
  if (w[3])
    strnzcpy (pNod[nNod].filebox, w[3], sizeof (pNod[nNod].filebox));
  else
    pNod[nNod].filebox[0] = 0;

  ++nNod;

  for (i = 0; i < 4; ++i)
    if (w[i])
      free (w[i]);
}

static int node_cmp (FTN_NODE *a, FTN_NODE *b)
{
  return ftnaddress_cmp (&a->fa, &b->fa);
}

FTN_NODE *get_node_info (FTN_ADDR *fa)
{
  FTN_NODE n;

  memcpy (&n.fa, fa, sizeof (FTN_ADDR));
  return (FTN_NODE *) bsearch (&n, pNod, nNod, sizeof (FTN_NODE),
			   (int (*) (const void *, const void *)) node_cmp);
}

FTN_DOMAIN *get_domain_info (char *domain_name)
{
  FTN_DOMAIN *curr;

  for (curr = pDomains; curr; curr = curr->next)
    if (!stricmp (curr->name, domain_name))
      return curr;
  return 0;
}

FTN_DOMAIN *get_def_domain ()
{
  FTN_DOMAIN *curr;

  for (curr = pDomains; curr->next; curr = curr->next);
  return curr;
}

static char *keys[] =
{
  "Log", "domain",
  "address", "sysname", "sysop", "location", "nodeinfo",
  "oport", "iport", "rescan-delay", "call-delay", "timeout", "oblksize",
  "inbound", "inbound-nonsecure", "Loglevel",
  "maxservers", "maxclients",
  "node",
  0
};

static meth methods[] =
{
  LogFile, Domain,
  Address, Sysname, Sysop, Location, Nodeinfo,
  Oport, Iport, RescanDelay, CallDelay, Timeout, OBlkSize,
  Inbound, InboundNonsecure, LogLevel,
  MaxServers, MaxClients,
  Node,
  0
};

#define TEST(var) if (!*var) Log (0, "%s: "#var" should be defined", path)

void readcfg (char *_path)
{
  FILE *in;
  char buf[MAXCFGLINE + 1];
  int i;

  line = 0;
  path = _path;

  if ((in = fopen (path, "r")) == 0)
    Log (0, "%s: %s", path, strerror (errno));

  while (!feof (in))
  {
    if (!fgets (buf, sizeof (buf), in))
      break;
    ++line;

    i = 0;
    while (buf[i] && isspace (buf[i]))
      ++i;
    if (buf[i] && buf[i] != '#')
    {
      char *s = strtok (buf + i, wsep);
      int j;

      if (s)
      {
	for (j = 0; keys[j]; ++j)
	  if (!stricmp (keys[j], s))
	    break;

	if (methods[j])
	{
	  methods[j] (strchr (s, 0) + 1);
	}
	else
	  Log (0, "%s: %i: %s: unknown keyword", path, line, s);
      }
    }
  }
  TEST (sysname);
  strwipe (sysname);
  TEST (sysop);
  strwipe (sysop);
  TEST (location);
  strwipe (location);
  TEST (nodeinfo);
  strwipe (nodeinfo);

  if (!*inbound_nonsecure)
    strcpy (inbound_nonsecure, inbound);

  if (!nAddr)
    Log (0, "%s: your address should be defined", path);

  if (pDomains == 0)
    Log (0, "%s: at least one domain should be defined", path);

  qsort (pNod, nNod, sizeof (FTN_NODE), (int (*) (const void *, const void *)) node_cmp);

  akas ();
}
